import System.IO
import System.Environment (getArgs)
import Data.List

data Geom      = Circle Point Double        -- (x , y)  radius
               | Line   Point Point         -- (x1,y1)  (x2,y2)
	deriving (Show,Eq,Ord)

type Fractal     = ([(Geom,Geom,Geom)],[Geom])--(objects in next generation,placed objects)
type Point       = (Double,Double)--(x,y)
type Vector      = (Double,Double)--(x,y)
type FractalData = (Double,(Vector,Vector,Vector))

getChildren::FractalData -> (Geom,Geom,Geom)->[(Geom,Geom,Geom)]   --linjer sist
getChildren (limit,tri) (a,b,c)
        | radius abc > limit = children
        | otherwise          = []
            where abc        = toCircle (a,b,c)
                  children   = [(a,b,abc),(neg a,c,abc),(neg b,neg c,abc),triToGeoms (placeTriangleInCircle abc tri)]

gen0::(Vector,Vector,Vector)->Fractal
gen0 (a,b,c) = ([( Line a b , Line a c , Line b c )],[]) 

nextGen::Fractal->FractalData->Fractal
nextGen (a,b) d =  (concat (map (getChildren d) a),(concat(map toGeoms a))++b)


genN::Integer->Fractal->FractalData->Fractal
genN 0 f      _   = f
genN _ ([],b) _   = ([],b)
genN x f      d   = nextGen (genN (x-1) f d) d

toGeoms g@(a,b,c)
     | all isLine [a,b,c] = [a,b,c,toCircle g]
     | otherwise          = [toCircle g]

--compile with "ghc -o foo fractal.hs"
--renderGenN n p1 p2 p3 limit
main = do
    s <- getArgs
    if length s /= 4 then 
         putStrLn "Please give parameters on the form: Generation (y1,x1) (y2,x2) (y3,x3) . Example: 12 \"(10,10)\" \"(600,900)\" \"(900,600)\" "
         else let [a,b,c,d] = s in
              renderGenN (read a) (read b) (read c) (read d)      
    


--geometry utilities
dist::Point->Point->Double
dist a b = len (a `diff` b)

diff,add::Vector->Vector->Vector
diff (x1,y1) (x2,y2) = (x1-x2,y1-y2) 
add (x1,y1) (x2,y2) = (x1+x2,y1+y2)

dot::Vector->Vector->Double
dot (x1,x2) (y1,y2) = x1*y1 + x2*y2

len::Vector->Double
len v = sqrt (v `dot` v)

v_scale::Vector->Double->Vector
v_scale (a,b) c = (c*a,c*b)

normalize::Vector->Vector
normalize v = v_scale v (1/(len v))

bisector::Vector->Vector->Vector
bisector u v = (v_scale u (len v)) `add` (v_scale v (len u))

n_bisector::Vector->Vector->Vector
n_bisector u v = normalize ( bisector u v)

neg::Geom->Geom
neg (Line a b) = Line b a 
neg x = x

proj::Vector->Vector->Vector
proj u v= v_scale v ((u `dot` v)/(v `dot` v))

lineIntersection (Line (x1,y1) (x2,y2)) (Line (x3,y3) (x4,y4)) = (x,y)
     where x  = ((x1*y2-y1*x2)*(x3-x4)-(x1-x2)*(x3*y4-y3*x4)) / ((x1-x2)*(y3-y4)-(y1-y2)*(x3-x4))
           y  = ((x1*y2-y1*x2)*(y3-y4)-(y1-y2)*(x3*y4-y3*x4)) / ((x1-x2)*(y3-y4)-(y1-y2)*(x3-x4))

circumcenter a b c = lineIntersection l1 l2
     where p1 = v_scale (a `add` b) (1/2)
           p2 = v_scale (a `add` c) (1/2)
           ac = c `diff` a
           ab = b `diff` a
           l1 = Line p1 (p1 `add` (ac `diff` (proj ac ab)))
           l2 = Line p2 (p2 `add` (ab `diff` (proj ab ac)))


circumradius a b c = a `dist` (circumcenter a b c)

--this function maps a triangle to a similar triangle with corners on the unit circle. 
normalizeTriangle::(Vector,Vector,Vector)->(Vector,Vector,Vector)
normalizeTriangle (a,b,c) = (f a,f b,f c) 
     where cs  = circumcenter a b c
           cr  = circumradius a b c
           f x = v_scale (x `diff` cs) (1/cr)

--assuming that (a,b,c) are points on the unit triangle, this function returns a similar triangle with corners on the given circle.
placeTriangleInCircle::Geom->(Vector,Vector,Vector)->(Vector,Vector,Vector)
placeTriangleInCircle cir (a,b,c) = (f a,f b,f c) 
     where f x = (v_scale x (radius cir)) `add` (center cir)


triToGeoms::(Vector,Vector,Vector)->(Geom,Geom,Geom)
triToGeoms (a,b,c) = (Line a b,Line a c,Line b c)


radius::Geom->Double
radius (Circle _ r) = r

center::Geom->Point
center (Circle c _) = c

isLine::Geom->Bool
isLine (Line _ _)   = True
isLine  _           = False

--Descartes theorem and related stuff

--This is a implementation of Descartes theorem, http://mathworld.wolfram.com/DescartesCircleTheorem.html .
descartes:: Geom->Geom->Geom->Double
descartes (Circle _ r1) (Circle _ r2) (Circle _ r3) = minimum [k x|x<-[pre+post,pre-post],x>=0.0]
      where k r  = (1/r)
            k1   = k r1
            k2   = k r2
            k3   = k r3
            pre  = k1 + k2 + k3
            post = 2*sqrt(k1*k2 + k1*k3 + k2*k3)

--Note that 1.0/0.0 = infinity, and a line is a circle with infinite radius.
descartes c1@(Circle _ _) c2@(Circle _ _) (Line _ _) = descartes c1 c2 (Circle (0.0,0.0) (1.0/0.0))


--This function finds the points of intersection between two circles, assuming that they intersect in exactly two points. 
findpoints (Circle c0@(x0,y0) r0) (Circle c1@(x1,y1) r1) = [(x2,y2) `add` rest,(x2,y2) `diff` rest]
      where d        =  c0 `dist` c1
            a        = (r0^2 - r1^2 + d^2 ) / (2*d)
            h        = sqrt (r0^2- a^2)
            (x2,y2)  = c0 `add` (v_scale (c1 `diff` c0) (a/d))
            rest     = v_scale (y1-y0,x0-x1) (h/d)

--Using Descartes theorem and findpoints, this function two possible candidates for placing the new circle.
getAllDescCircles::Geom->Geom->Geom->[Geom]
getAllDescCircles a@(Circle c0 r0) b@(Circle c1 r1) c = [Circle c r|c<-findpoints (Circle c0 (r0+r)) (Circle c1 (r1+r))] 
      where r = descartes a b c


--graphics functions

renderGenN 0 p1 p2 p3 = render [a,b,c] (getFileName 0 p1 p2 p3 limit)
    where (a,b,c) = triToGeoms (p1,p2,p3)
          limit   = 2.0
 
renderGenN n p1 p2 p3 = ((render.snd) $ genN n (gen0 (p1,p2,p3)) (limit,(normalizeTriangle (p1,p2,p3)))) (getFileName n p1 p2 p3 limit)
    where limit = 2.0


getFileName n p1 p2 p3 limit   = (concat ( intersperse "_" ["Fractal",(show n),(show p1),(show p2),(show p3),(show limit)]))++".svg"  

render::[Geom]->String->IO ()
render g file= do  
     inhandle  <- openFile "svghead.txt" ReadMode
     outhandle <- openFile file WriteMode 
     svgHead   <- hGetContents inhandle  
     hPutStrLn outhandle svgHead
     ((hPutStrLn outhandle).unlines) (map geomToSvg g)
     hPutStrLn outhandle "</svg>"
     hClose inhandle
     hFlush outhandle
     hClose outhandle

geomToSvg (Line (x1,y1) (x2,y2)) = concat ["     <line x1=\"",show x1,"\" y1=\"",show y1,"\" x2=\"",show x2,"\" y2=\"",show y2,"\" style=\"stroke:rgb(0,0,0);stroke-width:2\"/>"]

geomToSvg (Circle (x,y) r) = concat ["     <circle cx=\"",show x,"\" cy=\"",show y,"\" r=\"",show r,"\" stroke=\"black\" stroke-width=\"2\" fill=\"none\"/>"]


-- toCircle calculates the position of a new Circle, given the three geometrical objects that bounds it.

--this is a standard formula for the incircle.
toCircle::(Geom,Geom,Geom)->Geom
toCircle (Line a@(a1,a2) b@(b1,b2)  , Line c d  , Line e f@(c1,c2)  ) = Circle (x,y) r
	where
              s1      = a `dist` b
              s2      = c `dist` d
              s3      = e `dist` f
              (x1,y1) = b `diff` a
              (x2,y2) = d `diff` a
              per     = s1+s2+s3
              x       = (s3*a1 + s2*b1 + s1*c1) / per
              y       = (s3*a2 + s2*b2 + s1*c2) / per
              r       = (abs (x1*y2-x2*y1))     / per

-- (a,b) (a,c) (circle on bisector bac)
--using some linear algebra, it is relatively easy to place this circle.
toCircle (Line a@(a1,a2) b@(b1,b2) , Line a' c@(c1,c2) , Circle cc@(cx,cy) cr) = Circle c_out r_out
	where
		n_bis  = n_bisector (b `diff` a) (c `diff` a)
		bisproj= proj (n_bis) (b `diff` a)
		c_edge = a `add` (n_bis `v_scale` ( (cc `dist` a) - cr))
		top    = c_edge `dist` a 
		bot    = 1 + (n_bis `dist` bisproj)
		k      = top / bot
		c_out  = a `add` (v_scale n_bis k)
		r_out  = dist c_out c_edge 
		
--With Descartes circle thorem, the last two cases can be treated at the same time.
toCircle g@(g1,g2,g3) = bestmatch (getAllDescCircles g2 g3 g1) g

--given a list of candidates and three reference objects, bestmatch returns the candidate with the lowest sum of square distances to the references. 
bestmatch list (g1,g2,g3) = (snd.head.sort) [(sum [(distance x y)^2|y<-[g1,g2,g3]],x)|x<-list]


distance::Geom->Geom->Double
distance (Circle c1 r1) (Circle c2 r2) = abs ((dist c1 c2)  - (r1 + r2))

distance (Circle c r) (Line p1 p2)     = abs (((c `diff` p1) `dist` prj) - r)
      where prj = (c `diff` p1) `proj` (p2 `diff` p1)

